!=========================================================================
! Copyright (C) GemStone Systems, Inc. 1986-2009.  All Rights Reserved.
!
! $Id: ExceptionHandler.gs 22222 2009-09-02 00:36:20Z stever $
!
!=========================================================================

run
ExceptionHandler instVarNames = #( 
	#'gsException' 
	#'blockToUnwindForRetry' 
	#'blockToUnwindForResignal' 
	#'returnBlock' 
	#'resumeBlock' 
	#'exceptionSelector' 
	#'signaledException'
) ifFalse: [
"force new class to have no class history"
	Object subclass: 'ExceptionHandler'
	  instVarNames: #( 
		gsException
		blockToUnwindForRetry
		blockToUnwindForResignal
		returnBlock
		resumeBlock 
		exceptionSelector
		signaledException)
	  classVars: #( TraceLog)
	  classInstVars: #()
	  poolDictionaries: #[]
	  inDictionary: Globals
	  constraints: #[  ]
	  instancesInvariant: false
	  newVersionOf: nil
	  isModifiable: false
].
true.
%

removeallmethods ExceptionHandler
removeallclassmethods ExceptionHandler

category: 'For Documentation Installation only'
classmethod: ExceptionHandler
installDocumentation

| doc txt |
doc := GsClassDocumentation newForClass: self.

txt := (GsDocText new) details:
'This is an internal class that assists in the implementation of
ANSI Exceptions.'.
doc documentClassWith: txt.

self description: doc.
%

! ------------------- Class methods for ExceptionHandler
category: 'Trace Log'
classmethod: ExceptionHandler
disableTraceLog

	TraceLog := nil.
%
category: 'Trace Log'
classmethod: ExceptionHandler
enableTraceLog

	TraceLog := 0 -> (WriteStream on: String new).
%
category: 'Trace Log'
classmethod: ExceptionHandler
trace: aString

	self 
		trace: aString
		indent: 0.
%
category: 'Trace Log'
classmethod: ExceptionHandler
trace: aString indent: anInteger

	TraceLog ~~ nil ifTrue: [
		anInteger < 0 ifTrue: [TraceLog key: TraceLog key + anInteger].
		aString ~~ nil ifTrue: [
			TraceLog value cr.
			TraceLog key timesRepeat: [TraceLog value space].
			TraceLog value nextPutAll: aString.
		].
		anInteger > 0 ifTrue: [TraceLog key: TraceLog key + anInteger].
	].
%
category: 'Trace Log'
classmethod: ExceptionHandler
traceBegin: aString

	self 
		trace: aString
		indent: 2.
%
category: 'Trace Log'
classmethod: ExceptionHandler
traceBlock: aBlock

	| answer trace |
	self enableTraceLog.
	answer := aBlock value.
	trace := self traceLog.
	self disableTraceLog.
	^answer -> trace.
%
category: 'Trace Log'
classmethod: ExceptionHandler
traceEnd: aString

	self 
		trace: aString
		indent: -2.
%
category: 'Trace Log'
classmethod: ExceptionHandler
traceLog

	TraceLog == nil ifTrue: [^''].
	^TraceLog value contents.
%
! ------------------- Instance methods for ExceptionHandler
category: 'Exceptions'
method: ExceptionHandler
caughtEx: ex number: num cat: cat args: args handler: aBlock

	| originalCategory  newExceptionA nextException|
	resumeBlock := [:value | ^value].
	signaledException := Error
		category:	cat 
		number:		num 
		arguments:	args.
	signaledException handler: self.
	[
		originalCategory := ex category.
		ex category: IgnoredException.
		newExceptionA := self caughtExceptionWithAction: aBlock.
	] ensure: [
		ex category: originalCategory.
	].

	"We should only get here if we are to resignal..."
	nextException := Exception 
		_exceptionFor: newExceptionA gsNumber
		signalDictionary: newExceptionA gsCategory.
	nextException isNil ifTrue: [
		^newExceptionA _defaultAction.
	].
	^nextException 
		_signal: newExceptionA gsCategory 
		number: newExceptionA gsNumber 
		args: newExceptionA gsArguments.
%
category: 'Exceptions'
method: ExceptionHandler
caughtExceptionWithAction: aBlock

	blockToUnwindForResignal := [:newExceptionA | ^newExceptionA].
	returnBlock value: (aBlock value: signaledException).
	self error: 'Should not get here!'.
%
category: 'Exceptions'
method: ExceptionHandler
doTryBlock: aBlock

	| result |
	blockToUnwindForRetry := [:newBlock | ^newBlock].
	result := aBlock numberArgs = 1 
		ifTrue: [aBlock value: self]
		ifFalse: [aBlock value].
	self return: result.
%
category: 'Accessors'
method: ExceptionHandler
exceptionSelector

	^exceptionSelector.
%
category: 'Accessors'
method: ExceptionHandler
gsException

	^gsException.
%
category: 'Accessors'
method: ExceptionHandler
gsException: anException

	gsException := anException.
	exceptionSelector := ExceptionA.
%
category: 'Exceptions'
method: ExceptionHandler
isNested

	| nextEx |
	^(nextEx := self nextException) ~~ nil _and: [nextEx isStaticHandler not].
%
category: 'Exceptions'
method: ExceptionHandler
nextException

	^gsException
		_nextExceptionFor: signaledException gsNumber
		signalDictionary: signaledException gsCategory.
%
category: 'Exceptions'
method: ExceptionHandler
outer

	| nextException |
	(nextException := self nextException) == nil ifTrue: [
		^signaledException _defaultAction.
	].
	^nextException 
		_signal: signaledException 
		number: signaledException gsNumber 
		args: signaledException gsArguments.
%
category: 'Exceptions'
method: ExceptionHandler
pass

	self
		outer;
		signalEarlyReturn.
	self error: 'Should not get here!'.
%
category: 'Exceptions'
method: ExceptionHandler
resignalAs: anExceptionA

	blockToUnwindForResignal value: anExceptionA.
%
category: 'Exceptions'
method: ExceptionHandler
resume: anObject

	resumeBlock value: anObject.
%
category: 'Exceptions'
method: ExceptionHandler
retryUsing: aBlock

	blockToUnwindForRetry value: aBlock.
%
category: 'Exceptions'
method: ExceptionHandler
return: anObject

	returnBlock value: anObject.
%
category: 'Trace Log'
method: ExceptionHandler
trace: aString

	self class trace: 'ExceptionHandler(' , self asOop printString , ')>>' , aString.
%
category: 'Trace Log'
method: ExceptionHandler
traceBegin: aString

	self class traceBegin: aString.
%
category: 'Trace Log'
method: ExceptionHandler
traceEnd: aString

	self class traceEnd: aString.

%
category: 'Exceptions'
method: ExceptionHandler
try: aBlock on: anExceptionSelector do: anotherBlock

	| protectedBlock |
	protectedBlock := aBlock.
	exceptionSelector := anExceptionSelector.
	returnBlock := [:value | ^value].
	gsException := Exception 
		category: exceptionSelector errorCategoryToCatch
		number: exceptionSelector errorNumberToCatch 
		do: [:ex :cat :num :args | 
			"The following code can return from the #'on:do:' message by using the returnBlock"
			self caughtEx: ex number: num cat: cat args: args handler: anotherBlock.
			"If we fall through, the result of the above expression will be returned by the #'signal' message."
		].
	gsException handler: self.
	[true] whileTrue: [
		| newBlock |
		(newBlock := self doTryBlock: protectedBlock) notNil ifTrue: [protectedBlock := newBlock].
	].
%
