!=========================================================================
! Copyright (C) GemStone Systems, Inc. 1986-2009.  All Rights Reserved.
!
! $Id: ExceptionSet.gs 22222 2009-09-02 00:36:20Z stever $
!
!=========================================================================

removeallmethods ExceptionSet
removeallclassmethods ExceptionSet

category: 'For Documentation Installation only'
classmethod: ExceptionSet
installDocumentation

| doc txt |
doc := GsClassDocumentation newForClass: self.

txt := (GsDocText new) details:
'This protocol describes the behavior of objects that may be 
used to group a set of <exceptionSelector> objects into a 
single <exceptionSelector>. This is useful for establishing 
a single exception handler that may deal with several different 
types of exceptions.

Instances conform to exceptionSelector and Object.'.
doc documentClassWith: txt.

self description: doc.
%

! ------------------- Class methods for ExceptionSet
category: 'instance creation'
classmethod: ExceptionSet
new
	"Answer a new instance of the receiver."

	^self basicNew initialize
%
category: 'instance creation'
classmethod: ExceptionSet
with: selector1 with: selector2
	"Private - Answer a new instance of the receiver containing 
	the <exceptionSelector> arguments, selector1 and selector2"

	^self basicNew
		selectors: (OrderedCollection with: selector1 with: selector2)
%
! ------------------- Instance methods for ExceptionSet
category: 'ANSI - exceptionSelector'
method: ExceptionSet
, anotherException
	"Return an exception set that contains the receiver and the argument exception.
	This is commonly used to specify a set of exception selectors for an 
	'exception handler.'

	In addition to 'anotherException' the exception set that is returned contains
	all of the 'exception selectors' contained in the receiver.

	The returned object may or may not be the same object as the receiver."

	^anotherException exceptionSetWithExceptionSet: self.
%
category: 'ANSI support'
method: ExceptionSet
add: anException

	selectors add: anException.
%
category: 'ANSI support'
method: ExceptionSet
errorCategoryToCatch

	^self.
%
category: 'ANSI support'
method: ExceptionSet
errorNumberToCatch

	^nil.
%
category: 'ANSI support'
method: ExceptionSet
exceptionSetWithException: anException

	selectors add: anException.
%
category: 'ANSI support'
method: ExceptionSet
exceptionSetWithExceptionSet: anExceptionSet

	selectors addAll: anExceptionSet selectors.
%
category: 'ANSI - exceptionSelector'
method: ExceptionSet
handles: exception

	^selectors anySatisfy: [:es | es handles: exception]
%
category: 'ANSI support'
method: ExceptionSet
handlesCategory: anObject number: anInteger
	"This method may be overridden for ExceptionSelectors that are more discriminating.
	For example, one might want to trap all GemStoneError numbers in a certain range."

	^self handles: anObject.
%
category: 'initializing'
method: ExceptionSet
initialize

	selectors := OrderedCollection new.
%
category: 'accessing'
method: ExceptionSet
selectors
	"Answer the value of the receiver's instance variable 'selectors'.
	This method was automatically generated, but may be modified."

	^selectors
%
category: 'Exceptions'
method: ExceptionSet
selectors: anObject
	"Set the value of the receiver's instance variable 'selectors' to anObject.
	This method was automatically generated, but may be modified."

	selectors := anObject
%
category: 'ANSI support'
method: ExceptionSet
size 

	^selectors size.
%
