!=========================================================================
! Copyright (C) GemStone Systems, Inc. 1986-2009.  All Rights Reserved.
!
! $Id: ZeroDivide.gs 22222 2009-09-02 00:36:20Z stever $
!
!=========================================================================

removeallmethods ZeroDivide
removeallclassmethods ZeroDivide

category: 'For Documentation Installation only'
classmethod: ZeroDivide
installDocumentation

| doc txt |
doc := GsClassDocumentation newForClass: self.

txt := (GsDocText new) details:
'The class protocol describe the behavior of the global ZeroDivide. 
It is used to as an exception selector to catch zero divide exceptions 
and can also be used to signal that a division by zero error has 
occured. Zero divide exceptions are resumable so any message in this 
protocol that signal such an exception may ultimately return to their 
sender. The signaled exceptions generated by this type of object 
conform to the protocol <ZeroDivide>.

The instance protocol describes the behavior of exceptions that are 
signalled when an attempt is made to divide some number (the dividend) 
by zero.'.
doc documentClassWith: txt.

self description: doc.
%

! ------------------- Class methods for ZeroDivide
category: 'ANSI - exceptionInstantiator'
classmethod: ZeroDivide
dividend: argument
	"Signal the occurrence of a division by zero exception. Capture the number
	that was being divided such that it is available from the 'signaled exception.'

	If the message #dividend is subsequently sent to the <ZeroDivide> object
	that is the 'signaled exception' the value of 'argument' is returned."

	"It seems rare that this Exception would be signaled by sending this message.
	Typically, this Exception is signaled indirectly by the VM when a 
	division by zero occurs. So why not just take the usual path?!"

	argument / 0.
%
category: 'ANSI support'
classmethod: ZeroDivide
errorCategoryToCatch

	^GemStoneError.
%
category: 'ANSI support'
classmethod: ZeroDivide
errorNumberToCatch

	^self errorNumberToSignal.
%
category: 'ANSI support'
classmethod: ZeroDivide
errorNumberToSignal
	"NUM_ERR_INT_DIVISION_BY_ZERO	An attempt was made to divide an integer by zero.  
	Args: (1) the number"

	^2026.
%
category: 'ANSI - exceptionInstantiator'
classmethod: ZeroDivide
signal

	0 / 0.
%
! ------------------- Instance methods for ZeroDivide
category: 'ANSI - signaledException'
method: ZeroDivide
dividend

	^tag.
%
category: 'ANSI - signaledException'
method: ZeroDivide
isResumable
	"Division by zero is not considered fatal, so allow resumption."

	^true
%
category: 'ANSI - exceptionInstantiator'
method: ZeroDivide
signal

	0 / 0.
%
